/*
 * This one's for Linux
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <asm/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <linux/input.h>
#include <linux/ioctl.h>
#include <pthread.h>

#include "spinkit.h"

#define TRUE 1
#define FALSE 0

// Name of the device on our filesystem
static const char dev_name[] =  "/dev/input/event";
static char KitVersion[] = "SpinWarrior Kit V1.5";

// A mutex to protect the open call
static pthread_mutex_t device_mutex = PTHREAD_MUTEX_INITIALIZER;

typedef
  struct _SpinDevice_t
   {
    int dev_num;                   // numeric part of device name
    int fd;                        // the filedescriptor for the device
    struct timeval readTimeout;  // timeout for reading from the device
    ULONG readTimeoutVal;          // read timeout from SpinKitSetTimeout
    struct input_id info;
    char serial[20];
    SPINKIT_DATA lastValue;
   }
  SpinDevice_t;

static SpinDevice_t SpinWarriors[SPINKIT_MAX_DEVICES];
static ULONG numSpinWarriors = 0;

static void SpiniClear(void)
 {
  int i;

  for (i = 0; i < numSpinWarriors; i++)
   {
    if (SpinWarriors[i].fd != -1)
      close(SpinWarriors[i].fd);
   }
  memset(SpinWarriors, 0, sizeof(SpinWarriors));
  numSpinWarriors = 0;
 }

static SpinDevice_t *SpiniGetDeviceByHandle(SPINKIT_HANDLE iowHandle)
 {
  int i;

  for (i = 0; i < numSpinWarriors; i++)
    if ((SpinDevice_t *) iowHandle == &SpinWarriors[i])
      return &SpinWarriors[i];
  return NULL;
 }

SPINKIT_HANDLE SPINKIT_API SpinKitOpenDevice(void)
 {
  char ifname[80];
  int i;
  int fd;
  struct input_id device_info;

  pthread_mutex_lock(&device_mutex);
  SpiniClear();
  for(i = 0; i < SPINKIT_MAX_DEVICES; i++)
   {
    sprintf(ifname, "%s%d", dev_name, i);
    fd = open(ifname, O_RDONLY);
    if(fd != -1)
     {
      //now get the info for this device
      if(!ioctl(fd, EVIOCGID, &device_info))
       {
        if((device_info.bustype == BUS_USB) && (device_info.vendor == SPINKIT_VID) &&
          ((device_info.product == SPINKIT_PID24R4) || 
		   (device_info.product == SPINKIT_PID24R6) ||
           (device_info.product == SPINKIT_PID24A3) ||
		   (device_info.product == SPINKIT_PID28R4) || 
		   (device_info.product == SPINKIT_PID28R6) ||
           (device_info.product == SPINKIT_PID28A3)))
         {
          SpinWarriors[numSpinWarriors].dev_num = i;
          SpinWarriors[numSpinWarriors].info = device_info;
          SpinWarriors[numSpinWarriors].fd = fd;
          SpinWarriors[numSpinWarriors].readTimeoutVal = 0xFFFFFFFF;
          memset(SpinWarriors[numSpinWarriors].serial, 0, 20);
          ioctl(fd, EVIOCGUNIQ(20), SpinWarriors[numSpinWarriors].serial);
          numSpinWarriors++;
          fd = -1;
         }
       }
     }
    if (fd != -1)
      close(fd);
   }
  pthread_mutex_unlock(&device_mutex);
  return SpinKitGetDeviceHandle(1);
 }

ULONG SPINKIT_API SpinKitGetNumDevs(void)
 {
  return numSpinWarriors;
 }

SPINKIT_HANDLE SPINKIT_API SpinKitGetDeviceHandle(ULONG numDevice)
 {
  if (numDevice >= 1 && numDevice <= numSpinWarriors)
    return (SPINKIT_HANDLE) &SpinWarriors[numDevice - 1];
  else
    return NULL;
 }

BOOL SPINKIT_API SpinKitGetSerialNumber(SPINKIT_HANDLE iowHandle, PWCHAR serialNumber)
 {
  SpinDevice_t *dev;

  if (serialNumber == NULL)
    return FALSE;
  dev = SpiniGetDeviceByHandle(iowHandle);
  if(dev != NULL)
   {
    int i;

    for(i = 0; i < 9; i++)
      serialNumber[i] = (unsigned short) dev->serial[i];
   }
  else
    serialNumber[0] = 0;
  return serialNumber[0] != 0;
 }

ULONG SPINKIT_API SpinKitGetProductId(SPINKIT_HANDLE iowHandle)
 {
  SpinDevice_t *dev;

  dev = SpiniGetDeviceByHandle(iowHandle);
  if(dev != NULL)
    return (ULONG) dev->info.product;
  else
    return 0;
 }

ULONG SPINKIT_API SpinKitGetRevision(SPINKIT_HANDLE iowHandle)
 {
  SpinDevice_t *dev;

  dev = SpiniGetDeviceByHandle(iowHandle);
  if(dev != NULL)
    return (ULONG) dev->info.version;
  else
    return 0;
 }

void SPINKIT_API SpinKitCloseDevice(SPINKIT_HANDLE devHandle)
 { 
  pthread_mutex_lock(&device_mutex);
  SpiniClear();
  pthread_mutex_unlock(&device_mutex);
 }


BOOL SPINKIT_API SpinKitSetTimeout(SPINKIT_HANDLE devHandle, ULONG timeout)
 {
  SpinDevice_t *dev;

  dev = SpiniGetDeviceByHandle(devHandle);
  if(dev != NULL)
   {
    dev->readTimeout.tv_sec = (time_t)(timeout / 1000);
    dev->readTimeout.tv_usec = (long) ((timeout % 1000) * 1000);
    dev->readTimeoutVal = timeout;
   }
  return dev != NULL;
 }

BOOL SPINKIT_API SpinKitRead(SPINKIT_HANDLE devHandle, PSPINKIT_DATA SpinData)
 {
  SpinDevice_t *dev;
  struct timeval tv;
  fd_set rfds;
  int result;
  struct input_event ev;

  dev = SpiniGetDeviceByHandle(devHandle);
  if ((dev != NULL) && (SpinData != NULL))
    for (;;)
     {
      tv = dev->readTimeout;
      FD_ZERO(&rfds);      // clear the read-descriptor for select
      FD_SET(dev->fd, &rfds);   // add the filedescriptor to read-descriptor
      if (dev->readTimeoutVal != 0xFFFFFFFF)
        result = select(dev->fd + 1, &rfds, NULL, NULL, &tv);
      else
        result = select(dev->fd + 1, &rfds, NULL, NULL, NULL);
      if (result > 0 && FD_ISSET(dev->fd, &rfds))
        if (read(dev->fd, &ev, sizeof(ev)) == sizeof(ev))
         {
          if (ev.type != EV_SYN)
           {
            memset(SpinData, 0, SPINKIT_DATA_SIZE);
            switch (ev.type)
             {
              case EV_REL:
              case EV_ABS:
                SpinData->Spins[ev.code] = ev.value;
                break;
              case EV_KEY:
                SpinData->Buttons[ev.code & 0xF] = ev.value;
                break;
             }
            dev->lastValue = *SpinData;
            return TRUE;
           }
         }
        else
          return FALSE;
      else
      if(result == 0)
        return FALSE;
     }
  return FALSE;
 }

BOOL SPINKIT_API SpinKitReadNonBlocking(SPINKIT_HANDLE devHandle, PSPINKIT_DATA SpinData)
 {
  SpinDevice_t *dev;
  struct timeval tv;
  fd_set rfds;
  int result;
  struct input_event ev;

  dev = SpiniGetDeviceByHandle(devHandle);
  if(dev != NULL && SpinData != NULL)
   {
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    FD_ZERO(&rfds);            // clear the read-descriptor for select
    FD_SET(dev->fd, &rfds);  // add the filedescriptor to read-descriptor
    result = select(dev->fd + 1, &rfds, NULL, NULL, &tv);
    if (result > 0 && FD_ISSET(dev->fd, &rfds))
      if (read(dev->fd, &ev, sizeof(ev)) == sizeof(ev))
        if (ev.type != EV_SYN)
         {
          memset(SpinData, 0, SPINKIT_DATA_SIZE);
          switch (ev.type)
           {
            case EV_REL:
            case EV_ABS:
              SpinData->Spins[ev.code] = ev.value;
              break;
            case EV_KEY:
              SpinData->Buttons[ev.code & 0xF] = ev.value;
              break;
           }
          dev->lastValue = *SpinData;
          return TRUE;
         }
   }
  *SpinData = dev->lastValue;
  return FALSE;
 }

PCHAR SPINKIT_API SpinKitVersion(void)
 {
  return KitVersion;
 }
